@@ -1,47 +1,52 @@
 #!/usr/bin/env python
 
-"""Small wrapper library over paramiko that allows for parallel execution of SSH commands on remote hosts and executing simple single host commands over SSH.
+# This file is part of parallel-ssh.
+
+# Copyright (C) 2014 Panos Kittenis
+
+# This library is free software; you can redistribute it and/or
+# modify it under the terms of the GNU Lesser General Public
+# License as published by the Free Software Foundation, version 2.1.
+
+# This library is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# Lesser General Public License for more details.
+
+# You should have received a copy of the GNU Lesser General Public
+# License along with this library; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+
+"""Asynchronous parallel SSH library
 
 parallel-ssh uses asychronous network requests - there is *no* multi-threading or multi-processing used.
 
 This is a *requirement* for commands on many (hundreds/thousands/hundreds of thousands) of hosts which would grind a system to a halt simply by having so many processes/threads all wanting to execute if done with multi-threading/processing.
 
-See :mod:`pssh.SSHClient` and :mod:`pssh.ParallelSSHClient` for class documentation.
+The `libev event loop library<http://software.schmorp.de/pkg/libev.html>_` is utilised on *nix systems. Windows is not supported.
 
-Copyright (C) 2014 Panos Kittenis
-
-This library is free software; you can redistribute it and/or
-modify it under the terms of the GNU Lesser General Public
-License as published by the Free Software Foundation, version 2.1.
-
-This library is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-Lesser General Public License for more details.
-
-You should have received a copy of the GNU Lesser General Public
-License along with this library; if not, write to the Free Software
-Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+See :mod:`pssh.ParallelSSHClient` and :mod:`pssh.SSHClient` for class documentation.
 """
 
-import logging
-import gevent.pool
-from gevent import socket
 from gevent import monkey
 monkey.patch_all()
+import gevent.pool
+import warnings
+from socket import gaierror as sock_gaierror, error as sock_error
+import logging
 import paramiko
 import os
 
-host_logger = logging.getLogger('host_logging')
+host_logger = logging.getLogger('host_logger')
 handler = logging.StreamHandler()
 host_log_format = logging.Formatter('%(message)s')
 handler.setFormatter(host_log_format)
 host_logger.addHandler(handler)
 host_logger.setLevel(logging.INFO)
-NUM_RETRIES = 3
+DEFAULT_RETRIES = 3
 
 logger = logging.getLogger(__name__)
-    
+
 class UnknownHostException(Exception):
     """Raised when a host is unknown (dns failure)"""
     pass
@@ -57,8 +62,8 @@ class AuthenticationException(Exception):
     pass
 
 
-class ProxyCommandException(Exception):
-    """Raised on ProxyCommand error - ProxyCommand configured exited with error"""
+class SSHException(Exception):
+    """Raised on SSHException error - error authenticating with SSH server"""
     pass
 
 
@@ -70,10 +75,11 @@ class SSHClient(object):
     def __init__(self, host,
                  user=None, password=None, port=None,
                  pkey=None, forward_ssh_agent=True,
-                 _agent=None):
+                 num_retries=DEFAULT_RETRIES, _agent=None, timeout=None,
+                 proxy_host=None, proxy_port=22):
         """Connect to host honouring any user set configuration in ~/.ssh/config \
         or /etc/ssh/ssh_config
-         
+        
         :param host: Hostname to connect to
         :type host: str
         :param user: (Optional) User to login as. Defaults to logged in user or \
@@ -87,6 +93,12 @@ class SSHClient(object):
         :type port: int
         :param pkey: (Optional) Client's private key to be used to connect with
         :type pkey: :mod:`paramiko.PKey`
+        :param num_retries: (Optional) Number of retries for connection attempts\
+        before the client gives up. Defaults to 3.
+        :type num_retries: int
+        :param timeout: (Optional) Number of seconds to timout connection attempts\
+        before the client gives up. Defaults to 10.
+        :type timeout: int
         :param forward_ssh_agent: (Optional) Turn on SSH agent forwarding - \
         equivalent to `ssh -A` from the `ssh` command line utility. \
         Defaults to True if not set.
@@ -96,16 +108,18 @@ class SSHClient(object):
         connecting to local SSH agent to lookup keys with our own SSH agent. \
         Only really useful for testing, hence the internal variable prefix.
         :type _agent: :mod:`paramiko.agent.Agent`
-        :raises: :mod:`pssh.AuthenticationException` on authentication error
-        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
-        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
-        :raises: :mod:`pssh.ProxyCommandException` on error with ProxyCommand configured
+        :param proxy_host: (Optional) SSH host to tunnel connection through
+        so that SSH clients connects to self.host via client -> proxy_host -> host
+        :type proxy_host: str
+        :param proxy_port: (Optional) SSH port to use to login to proxy host if set.
+        Defaults to 22.
+        :type proxy_port: int
         """
         ssh_config = paramiko.SSHConfig()
         _ssh_config_file = os.path.sep.join([os.path.expanduser('~'),
                                              '.ssh',
                                              'config'])
-        # Load ~/.ssh/config if it exists to pick up username, ProxyCommand
+        # Load ~/.ssh/config if it exists to pick up username
         # and host address if set
         if os.path.isfile(_ssh_config_file):
             ssh_config.parse(open(_ssh_config_file))
@@ -119,55 +133,86 @@ class SSHClient(object):
         client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
         self.forward_ssh_agent = forward_ssh_agent
         self.client = client
-        self.channel = None
         self.user = user
         self.password = password
         self.pkey = pkey
         self.port = port if port else 22
         self.host = resolved_address
-        self.proxy_command = paramiko.ProxyCommand(host_config['proxycommand']) if 'proxycommand' in \
-                             host_config else None
-        if self.proxy_command:
-            logger.debug("Proxy configured for destination host %s - ProxyCommand: '%s'",
-                         self.host, " ".join(self.proxy_command.cmd),)
         if _agent:
             self.client._agent = _agent
-        self._connect()
+        self.num_retries = num_retries
+        self.timeout = timeout
+        self.proxy_host, self.proxy_port = proxy_host, proxy_port
+        self.proxy_client = None
+        if self.proxy_host and self.proxy_port:
+            logger.debug("Proxy configured for destination host %s - Proxy host: %s:%s",
+                         self.host, self.proxy_host, self.proxy_port,)
+            self._connect_tunnel()
+        else:
+            self._connect(self.client, self.host, self.port)
 
-    def _connect(self, retries=1):
-        """Connect to host, throw UnknownHost exception on DNS errors"""
+    def _connect_tunnel(self):
+        """Connects to SSH server via an intermediate SSH tunnel server.
+        client (me) -> tunnel (ssh server to proxy through) -> \
+        destination (ssh server to run command)
+        :rtype: `:mod:paramiko.SSHClient` Client to remote SSH destination
+        via intermediate SSH tunnel server."""
+        self.proxy_client = paramiko.SSHClient()
+        self.proxy_client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
+        self._connect(self.proxy_client, self.proxy_host, self.proxy_port)
+        logger.info("Connecting via SSH proxy %s:%s -> %s:%s", self.proxy_host,
+                    self.proxy_port, self.host, self.port,)
+        proxy_channel = self.proxy_client.get_transport().\
+          open_channel('direct-tcpip', (self.host, self.port,),
+                       ('127.0.0.1', 0))
+        return self._connect(self.client, self.host, self.port, sock=proxy_channel)
+        
+    def _connect(self, client, host, port, sock=None, retries=1):
+        """Connect to host
+        
+        :raises: :mod:`pssh.AuthenticationException` on authentication error
+        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
+        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
+        :raises: :mod:`pssh.SSHException` on other undefined SSH errors
+        """
         try:
-            self.client.connect(self.host, username=self.user,
-                                password=self.password, port=self.port,
-                                pkey=self.pkey,
-                                sock=self.proxy_command)
-        except socket.gaierror, e:
+            client.connect(host, username=self.user,
+                           password=self.password, port=port,
+                           pkey=self.pkey,
+                           sock=sock, timeout=self.timeout)
+        except sock_gaierror, ex:
             logger.error("Could not resolve host '%s' - retry %s/%s",
-                         self.host, retries, NUM_RETRIES)
-            while retries < NUM_RETRIES:
+                         self.host, retries, self.num_retries)
+            while retries < self.num_retries:
                 gevent.sleep(5)
-                return self._connect(retries=retries+1)
-            raise UnknownHostException("%s - %s" % (str(e.args[1]),
-                                                    self.host,))
-        except socket.error, e:
+                return self._connect(client, host, port, sock=sock,
+                                     retries=retries+1)
+            raise UnknownHostException("%s - %s - retry %s/%s",
+                                       str(ex.args[1]),
+                                       self.host, retries, self.num_retries)
+        except sock_error, ex:
             logger.error("Error connecting to host '%s:%s' - retry %s/%s",
-                         self.host, self.port, retries, NUM_RETRIES)
-            while retries < NUM_RETRIES:
+                         self.host, self.port, retries, self.num_retries)
+            while retries < self.num_retries:
                 gevent.sleep(5)
-                return self._connect(retries=retries+1)
+                return self._connect(client, host, port, sock=sock,
+                                     retries=retries+1)
+            error_type = ex.args[1] if len(ex.args) > 1 else ex.args[0]
             raise ConnectionErrorException("%s for host '%s:%s' - retry %s/%s",
-                                           str(e.args[1]), self.host, self.port,
-                                           retries, NUM_RETRIES,)
-        except paramiko.AuthenticationException, e:
-            raise AuthenticationException(e)
-        except paramiko.ProxyCommandFailure, e:
-            logger.error("Error executing ProxyCommand - %s", e.message,)
-            raise ProxyCommandException(e.message)
+                                           str(error_type), self.host, self.port,
+                                           retries, self.num_retries,)
+        except paramiko.AuthenticationException, ex:
+            raise AuthenticationException(ex)
+        # SSHException is more general so should be below other types
+        # of SSH failure
+        except paramiko.SSHException, ex:
+            logger.error("General SSH error - %s", ex)
+            raise SSHException(ex)
 
-    def exec_command(self, command, sudo=False, **kwargs):
+    def exec_command(self, command, sudo=False, user=None, **kwargs):
         """Wrapper to :mod:`paramiko.SSHClient.exec_command`
 
-        Opens a new SSH session with a pty and runs command with given \
+        Opens a new SSH session with a new pty and runs command with given \
         `kwargs` if any. Greenlet then yields (sleeps) while waiting for \
         command to finish executing or channel to close indicating the same.
 
@@ -187,20 +232,34 @@ class SSHClient(object):
         if self.forward_ssh_agent:
             agent_handler = paramiko.agent.AgentRequestHandler(channel)
         channel.get_pty()
-        # stdin (unused), stdout, stderr
-        (_, stdout, stderr) = (channel.makefile('wb'), channel.makefile('rb'),
-                               channel.makefile_stderr('rb'))
-        if sudo:
+        if self.timeout:
+            channel.settimeout(self.timeout)
+        _stdout, _stderr = channel.makefile('rb'), \
+                           channel.makefile_stderr('rb')
+        stdout, stderr = self._read_output_buffer(_stdout,), \
+                         self._read_output_buffer(_stderr,
+                                                  prefix='\t[err]')
+        if sudo and not user:
             command = 'sudo -S bash -c "%s"' % command.replace('"', '\\"')
+        elif user:
+            command = 'sudo -u %s -S bash -c "%s"' % (
+                user, command.replace('"', '\\"'),)
         else:
             command = 'bash -c "%s"' % command.replace('"', '\\"')
         logger.debug("Running command %s on %s", command, self.host)
         channel.exec_command(command, **kwargs)
-        logger.debug("Command finished executing")
+        logger.debug("Command started")
         while not (channel.recv_ready() or channel.closed):
             gevent.sleep(.2)
         return channel, self.host, stdout, stderr
 
+    def _read_output_buffer(self, output_buffer, prefix=''):
+        """Read from output buffers and log to host_logger"""
+        for line in output_buffer:
+            output = line.strip()
+            host_logger.info("[%s]%s\t%s", self.host, prefix, output,)
+            yield output
+
     def _make_sftp(self):
         """Make SFTP client from open transport"""
         transport = self.client.get_transport()
@@ -255,78 +314,81 @@ class ParallelSSHClient(object):
     """Uses :mod:`pssh.SSHClient`, performs tasks over SSH on multiple hosts in \
     parallel.
 
-    Connections to hosts are established in parallel when ``exec_command`` is called,
+    Connections to hosts are established in parallel when ``run_command`` is called,
     therefor any connection and/or authentication exceptions will happen on the
-    call to ``exec_command`` and need to be caught."""
+    call to ``run_command`` and need to be caught."""
 
     def __init__(self, hosts,
                  user=None, password=None, port=None, pkey=None,
-                 forward_ssh_agent=True,
-                 pool_size=10):
+                 forward_ssh_agent=True, num_retries=DEFAULT_RETRIES, timeout=None,
+                 pool_size=10, proxy_host=None, proxy_port=22):
         """
         :param hosts: Hosts to connect to
         :type hosts: list(str)
-        :param user: (Optional) User to login as. Defaults to logged in user or\
+        :param user: (Optional) User to login as. Defaults to logged in user or \
         user from ~/.ssh/config or /etc/ssh/ssh_config if set
         :type user: str
-        :param password: (Optional) Password to use for login. Defaults to\
+        :param password: (Optional) Password to use for login. Defaults to \
         no password
         :type password: str
-        :param port: (Optional) Port number to use for SSH connection. Defaults\
+        :param port: (Optional) Port number to use for SSH connection. Defaults \
         to None which uses SSH default
         :type port: int
         :param pkey: (Optional) Client's private key to be used to connect with
         :type pkey: :mod:`paramiko.PKey`
+        :param num_retries: (Optional) Number of retries for connection attempts \
+        before the client gives up. Defaults to 3.
+        :type num_retries: int
+        :param timeout: (Optional) Number of seconds to timout connection \
+        attempts before the client gives up. Defaults to 10.
+        :type timeout: int
         :param forward_ssh_agent: (Optional) Turn on SSH agent forwarding - \
         equivalent to `ssh -A` from the `ssh` command line utility. \
         Defaults to True if not set.
         :type forward_ssh_agent: bool
         :param pool_size: (Optional) Greenlet pool size. Controls on how many\
-        hosts to execute tasks in parallel. Defaults to 10
+        hosts to execute tasks in parallel. Defaults to number of hosts or 10, \
+        whichever is lower. Pool size will be *equal to* number of hosts if number\
+        of hosts is lower than the pool size specified as that would only \
+        increase overhead with no benefits.
         :type pool_size: int
-        :raises: :mod:`pssh.AuthenticationException` on authentication error
-        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
-        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
-        :raises: :mod:`pssh.ProxyCommandException` on error with ProxyCommand configured
+        :param proxy_host: (Optional) SSH host to tunnel connection through
+        so that SSH clients connects to self.host via client -> proxy_host -> host
+        :type proxy_host: str
+        :param proxy_port: (Optional) SSH port to use to login to proxy host if set.
+        Defaults to 22.
+        :type proxy_port: int
         
         **Example**
 
-        >>> from pssh import ParallelSSHClient, AuthenticationException,\
-        		UnknownHostException, ConnectionErrorException
+        >>> from pssh import ParallelSSHClient, AuthenticationException, \
+UnknownHostException, ConnectionErrorException
         >>> client = ParallelSSHClient(['myhost1', 'myhost2'])
         >>> try:
-        >>> ... cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
+        >>> ... output = client.run_command('ls -ltrh /tmp/aasdfasdf', sudo=True)
         >>> except (AuthenticationException, UnknownHostException, ConnectionErrorException):
         >>> ... return
-        >>> output = [client.get_stdout(cmd) for cmd in cmds]
+        >>> # Commands have started executing at this point
+        >>> # Exit code will probably not be available immediately
+        >>> print output
+        >>> {'myhost1': {'exit_code': None,
+                         'stdout' : <generator>,
+                         'stderr' : <generator>,
+                         'cmd' : <greenlet>,
+                         },
+             'myhost2': {'exit_code': None,
+                         'stdout' : <generator>,
+                         'stderr' : <generator>,
+                         'cmd' : <greenlet>,
+            }}
+        >>> # Print output as it comes in. 
+        >>> for host in output: for line in output[host]['stdout']: print line
         [myhost1]     ls: cannot access /tmp/aasdfasdf: No such file or directory
         [myhost2]     ls: cannot access /tmp/aasdfasdf: No such file or directory
-        >>> print output
-        [{'myhost1': {'exit_code': 2}}, {'myhost2': {'exit_code': 2}}]
-
-        **Example with returned stdout and stderr buffers**
-
-        >>> from pssh import ParallelSSHClient, AuthenticationException,\
-        		UnknownHostException, ConnectionErrorException
-        >>> client = ParallelSSHClient(['myhost1', 'myhost2'])
-        >>> try:
-        >>> ... cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf', sudo = True)
-        >>> except (AuthenticationException, UnknownHostException, ConnectionErrorException):
-        >>> ... return
-        >>> output = [client.get_stdout(cmd, return_buffers=True) for cmd in cmds]
-        >>> print output
-        [{'myhost1': {'exit_code': 2,
-        	      'stdout' : <generator object <genexpr>,
-                      'stderr' : <generator object <genexpr>,}},
-         {'myhost2': {'exit_code': 2,
-         	      'stdout' : <generator object <genexpr>,
-                      'stderr' : <generator object <genexpr>,}},
-                      ]
-        >>> for host_stdout in output:
-            ... for line in host_stdout[host_output.keys()[0]]['stdout']:
-                ... print line
-                    ls: cannot access /tmp/aasdfasdf: No such file or directory
-                    ls: cannot access /tmp/aasdfasdf: No such file or directory
+        >>> # Retrieve exit code after commands have finished
+        >>> # `get_exit_code` will return `None` if command has not finished
+        >>> print client.get_exit_code(output[host])
+        0
 
         **Example with specified private key**
 
@@ -335,15 +397,15 @@ class ParallelSSHClient(object):
         >>> client = ParallelSSHClient(['myhost1', 'myhost2'], pkey=client_key)
         
         .. note ::
-          
+        
           **Connection persistence**
           
           Connections to hosts will remain established for the duration of the
           object's life. To close them, just `del` or reuse the object reference.
           
           >>> client = ParallelSSHClient(['localhost'])
-          >>> cmds = client.exec_command('ls -ltrh /tmp/aasdfasdf')
-          >>> cmds[0].join()
+          >>> output = client.run_command('ls -ltrh /tmp/aasdfasdf')
+          >>> client.pool.join()
           
           :netstat: ``tcp        0      0 127.0.0.1:53054         127.0.0.1:22            ESTABLISHED``
           
@@ -354,20 +416,85 @@ class ParallelSSHClient(object):
           
           Connection is terminated.
         """
-        self.pool = gevent.pool.Pool(size=pool_size)
-        self.pool_size = pool_size
+        self.pool_size = len(hosts) if len(hosts) < pool_size else pool_size
+        self.pool = gevent.pool.Pool(size=self.pool_size)
         self.hosts = hosts
         self.user = user
         self.password = password
         self.forward_ssh_agent = forward_ssh_agent
         self.port = port
         self.pkey = pkey
+        self.num_retries = num_retries
+        self.timeout = timeout
+        self.proxy_host, self.proxy_port = proxy_host, proxy_port
         # To hold host clients
         self.host_clients = dict((host, None) for host in hosts)
 
+    def run_command(self, *args, **kwargs):
+        """Run command on all hosts in parallel, honoring self.pool_size,
+        and return output buffers.
+
+        This function will block until all commands have **started** and
+        then return immediately. Any connection and/or authentication exceptions
+         will be raised here and need catching.
+
+        :param args: Positional arguments for command
+        :type args: tuple
+        :param kwargs: Keyword arguments for command
+        :type kwargs: dict
+
+        :rtype: Dictionary with host as key as per :mod:`ParallelSSH.get_output`:
+        
+        :raises: :mod:`pssh.AuthenticationException` on authentication error
+        :raises: :mod:`pssh.UnknownHostException` on DNS resolution error
+        :raises: :mod:`pssh.ConnectionErrorException` on error connecting
+        :raises: :mod:`pssh.SSHException` on other undefined SSH errors
+        
+        ::
+        
+          {'myhost1': {'exit_code': exit code if ready else None,
+                       'channel' : SSH channel of command,
+                       'stdout'  : <iterable>,
+                       'stderr'  : <iterable>,
+                       'cmd'     : <greenlet>}}
+        
+        **Example**:
+        
+        >>> output = client.run_command('ls -ltrh')
+
+        print stdout for each command:
+        
+        >>> for host in output:
+        >>>     for line in output[host]['stdout']: print line
+
+        Get exit code after command has finished:
+
+        >>> for host in output:
+        >>>     for line in output[host]['stdout']: print line
+        >>>     exit_code = client.get_exit_code(output[host])
+        
+        Wait for completion, no stdout:
+        
+        >>> client.pool.join()
+        
+        Capture stdout - **WARNING** - this will store the entirety of stdout
+        into memory and may exhaust available memory if command output is
+        large enough:
+        
+        >>> for host in output:
+        >>>     stdout = list(output[host]['stdout'])
+        >>>     print "Complete stdout for host %s is %s" % (host, stdout,)
+        
+        """
+        for host in self.hosts:
+            self.pool.spawn(self._exec_command, host, *args, **kwargs)
+        return self.get_output()
+    
     def exec_command(self, *args, **kwargs):
         """Run command on all hosts in parallel, honoring self.pool_size
 
+        **Superseeded by :mod:`ParallelSSH.run_command`**
+
         :param args: Position arguments for command
         :type args: tuple
         :param kwargs: Keyword arguments for command
@@ -376,7 +503,7 @@ class ParallelSSHClient(object):
         :rtype: List of :mod:`gevent.Greenlet`
 
         **Example**:
-      
+        
         >>> cmds = client.exec_command('ls -ltrh')
         
         Wait for completion, no stdout:
@@ -393,50 +520,111 @@ class ParallelSSHClient(object):
         
         You may call get_stdout on already completed greenlets to re-get
         their output as many times as you want."""
+        warnings.warn("This method is being deprecated and will be removed in \
+future releases - use self.run_command instead", DeprecationWarning)
         return [self.pool.spawn(self._exec_command, host, *args, **kwargs)
                 for host in self.hosts]
-    
+
     def _exec_command(self, host, *args, **kwargs):
         """Make SSHClient, run command on host"""
         if not self.host_clients[host]:
             self.host_clients[host] = SSHClient(host, user=self.user,
                                                 password=self.password,
                                                 port=self.port, pkey=self.pkey,
-                                                forward_ssh_agent=self.forward_ssh_agent)
+                                                forward_ssh_agent=self.forward_ssh_agent,
+                                                num_retries=self.num_retries,
+                                                timeout=self.timeout,
+                                                proxy_host=self.proxy_host,
+                                                proxy_port=self.proxy_port)
         return self.host_clients[host].exec_command(*args, **kwargs)
 
+    def get_output(self, commands=None):
+        """Get output from running commands.
+
+        Stdout and stderr are also logged via the logger named ``host_logger``
+        which is enabled by default.
+        ``host_logger`` output can be disabled by removing its handler.
+        
+        >>> logger = logging.getLogger('host_logger')
+        >>> for handler in logger.handlers: logger.removeHandler(handler)
+        
+        **Example usage**:
+
+        >>> output = client.get_output()
+        >>> for host in output: print output[host]['stdout']
+        <stdout>
+        >>> # Get exit code after command has finished
+        >>> self.get_exit_code(output[host])
+        0
+
+        :param commands: (Optional) Override commands to get output from.
+        Uses running commands in pool if not given
+        :type commands: :mod:`gevent.Greenlet`
+        :rtype: Dictionary with host as key as in:
+
+        ::
+        
+          {'myhost1': {'exit_code': exit code if ready else None,
+                       'channel' : SSH channel of command,
+                       'stdout'  : <iterable>,
+                       'stderr'  : <iterable>,
+                       'cmd'     : <greenlet>}}
+        
+        """
+        if not commands:
+            commands = list(self.pool.greenlets)
+        output = {}
+        for cmd in commands:
+            (channel, host, stdout, stderr) = cmd.get()
+            output.setdefault(host, {})
+            output[host].update({'exit_code': self._get_exit_code(channel),
+                                 'channel' : channel,
+                                 'stdout' : stdout,
+                                 'stderr' : stderr,
+                                 'cmd' : cmd, })
+        return output
+
+    def get_exit_code(self, host_output):
+        """Get exit code from host output if available
+        :param host_output: Per host output as returned by `self.get_output`
+        :rtype: int or None if exit code not ready"""
+        if not 'channel' in host_output:
+            logger.error("%s does not look like host output..", host_output,)
+            return
+        channel = host_output['channel']
+        return self._get_exit_code(channel)
+
+    def _get_exit_code(self, channel):
+        """Get exit code from channel if ready"""
+        if not channel.exit_status_ready():
+            return
+        channel.close()
+        return channel.recv_exit_status()
+
     def get_stdout(self, greenlet, return_buffers=False):
         """Get/print stdout from greenlet and return exit code for host
         
-        :mod:`pssh.get_stdout` will close the open SSH channel but this does
-        **not** close the established connection to the remote host, only the
-        authenticated SSH channel within it. This is standard practise
-        in SSH when a command has finished executing. A new command
-        will open a new channel which is very fast on already established
-        connections.
-
-        By default, stdout and stderr will be logged via the logger named \
-        ``host_logger`` unless ``return_buffers`` is set to ``True`` in which case \
-        both buffers are instead returned along with the exit status.
-
+        **Deprecated** - use self.get_output() instead.
+        
         :param greenlet: Greenlet object containing an \
         SSH channel reference, hostname, stdout and stderr buffers
         :type greenlet: :mod:`gevent.Greenlet`
-
+        
         :param return_buffers: Flag to turn on returning stdout and stderr \
         buffers along with exit code. Defaults to off.
         :type return_buffers: bool
-
+        
         :rtype: Dictionary containing ``{host: {'exit_code': exit code}}`` entry \
         for example ``{'myhost1': {'exit_code': 0}}``
         :rtype: With ``return_buffers=True``: ``{'myhost1': {'exit_code': 0,
                                                              'channel' : None or SSH channel of command if command is still executing,
-         						     'stdout' : <iterable>,
+                                                             'stdout' : <iterable>,
                                                              'stderr' : <iterable>,}}``
         """
-        channel, host, _stdout, _stderr = greenlet.get()
-        stdout = self._read_output_buffer(_stdout)
-        stderr = self._read_output_buffer(_stderr)
+        warnings.warn("This method is being deprecated and will be removed in \
+future releases - use self.get_output instead", DeprecationWarning)
+        gevent.sleep(.2)
+        channel, host, stdout, stderr = greenlet.get()
         if channel.exit_status_ready():
             channel.close()
         else:
@@ -445,10 +633,10 @@ class ParallelSSHClient(object):
             return_buffers = True
         # Channel must be closed or reading stdout/stderr will block forever
         if not return_buffers and channel.closed:
-            for line in stdout:
-                host_logger.info("[%s]\t%s", host, line,)
-            for line in stderr:
-                host_logger.info("[%s] [err] %s", host, line,)
+            for _ in stdout:
+                pass
+            for _ in stderr:
+                pass
             return {host: {'exit_code': channel.recv_exit_status(),}}
         gevent.sleep(.2)
         return {host: {'exit_code': channel.recv_exit_status() if channel.exit_status_ready() else None,
@@ -456,22 +644,6 @@ class ParallelSSHClient(object):
                        'stdout' : stdout,
                        'stderr' : stderr, }}
 
-    def wait_for_exit_status(self, channel):
-        """Block and wait for exit status on channel.
-        WARNING - this will block forever if the command executed never exits
-        :rtype: int - Exit code of command executed"""
-        while not channel.exit_status_ready():
-            gevent.sleep(1)
-        channel.close()
-        return channel.recv_exit_status()
-
-    def _read_output_buffer(self, output_buffer):
-        """Read from output buffers,
-        allowing coroutines to execute in between reading"""
-        for line in output_buffer:
-            gevent.sleep(1)
-            yield line.strip()
-
     def copy_file(self, local_file, remote_file):
         """Copy local file to remote file in parallel
         
